#ifndef FERI_ROUTINES
#define FERI_ROUTINES
// -----------------------------------------------
// ATTRACTOR ANALYSIS WITH GRAPH REPRESENTATION
//
// Main Routines

// -----------------------------------------------

#include "main.hpp"
#include "typedefs.hpp"

// -----------------------------------------------

// evaluating Enclosure
bool evaluateEnclosure(double &boxSize, const GraphType &GRAPH, VectorType &box)
{
	std::pair<VertexIterator, VertexIterator> GRAPHlimits;
	VertexIterator it;
	VectorType tempbox;


	if (num_vertices(GRAPH) == 0) {

		return false;

	} else {

		GRAPHlimits = boost::vertices(GRAPH);
		it = GRAPHlimits.first;
		box = GRAPH[*it].encl;

		boxSize = 0;

		while ( it != GRAPHlimits.second ) {

			tempbox = GRAPH[*it].encl;
			for (int i = 1; i <= box.dimension(); ++i) {
				boxSize = capd::max(boxSize, diam(tempbox(i)).leftBound());
				box(i) = intervalHull(box(i), tempbox(i));
			}
			++it;

		}

		return true;

	}

}

// plotting the boxes (i,j)-plot
void plotBoxes(const GraphType &GRAPH, const int &i, const int &j)
{
	std::pair<VertexIterator, VertexIterator> GRAPHlimits;
	VertexIterator it;

	double lt1, lt2, rb1, rb2;



	GRAPHlimits = boost::vertices(GRAPH);
	it = GRAPHlimits.first;

	while ( it != GRAPHlimits.second ) {

		lt1 = GRAPH[*it].encl(i).leftBound();
		lt2 = GRAPH[*it].encl(j).leftBound();
		rb1 = GRAPH[*it].encl(i).rightBound();
		rb2 = GRAPH[*it].encl(j).rightBound();

		rootFrame.boxFill(lt1, lt2, rb1, rb2, BLUE);

		++it;

	}

	waitBt();
}

// building initial boxes
void buildBoxes(const VectorType &V, VertexType &vbox, const double &initialBoxSize, GraphType &GRAPH, const int component)
{
	double x1, x2;

	x1 = V(component).leftBound();
	x2 = x1 + initialBoxSize;

	if (component == 1) {

		while (x1 < V(1).rightBound()) {
			vbox.encl(1) = interval(x1, x2);
			boost::add_vertex(vbox, GRAPH);

			x1 = x2;
			x2 = x2 + initialBoxSize;
		}

	} else {

		while (x1 < V(component).rightBound()) {
			vbox.encl(component) = interval(x1, x2);
			buildBoxes(V, vbox, initialBoxSize, GRAPH, component - 1);

			x1 = x2;
			x2 = x2 + initialBoxSize;
		}

	}
}

// building initial boxes
void buildInitialBoxes(const VectorType &V, const double &initialBoxSize, GraphType &GRAPH)
{
	VertexType vbox;

	vbox.visited = false;
	buildBoxes(V, vbox, initialBoxSize, GRAPH, V.dimension());
}

// adding initial edges
template<typename VFType>
void addInitialEdges(VFType &TheMap, GraphType &GRAPH)
{
	std::pair<VertexIterator, VertexIterator> GRAPHlimits;
	VertexIterator it, it2;

	VectorType img, tempbox;



	GRAPHlimits = boost::vertices(GRAPH);
	it = GRAPHlimits.first;

	while ( it != GRAPHlimits.second ) {

		it2 = GRAPHlimits.first;
		TheMap.evaluate(GRAPH[*it].encl, img);

		while ( it2 != GRAPHlimits.second ) {

			if ( intersection(GRAPH[*it2].encl, img, tempbox) )
				boost::add_edge(*it, *it2, GRAPH);

			++it2;

		}

		++it;

	}
}

#ifdef FERI_TARJAN
// find the current strongly connected component [Tarjan alg., Recursive Version]
void StrongConnect(GraphType &GRAPH, VertexStack &S, long int &index, VertexDescriptor v)
{

	std::pair<outEdgeIterator, outEdgeIterator> outEdges;
	outEdgeIterator oute;

	VertexDescriptor w;

	bool self_connected;

	self_connected = false;



	GRAPH[v].tarjan_index = index;
	GRAPH[v].tarjan_lowlink = index;

	index += 1;

	S.push(v);
	GRAPH[v].tarjan_inS = true;


	outEdges = boost::out_edges(v, GRAPH);
	oute = outEdges.first;
	while ( oute != outEdges.second ) {
		w = target(*oute, GRAPH);

		// there is a self-connection
		if (w == v) {
			self_connected = true;
		}

		// new node
		if (GRAPH[w].tarjan_index == -1) {

			StrongConnect(GRAPH, S, index, w);
			GRAPH[v].tarjan_lowlink = std::min(GRAPH[v].tarjan_lowlink, GRAPH[w].tarjan_lowlink);

		// not new node but already in S
		} else if (GRAPH[w].tarjan_inS) {

			GRAPH[v].tarjan_lowlink = std::min(GRAPH[v].tarjan_lowlink, GRAPH[w].tarjan_lowlink);

		}

		++oute;
	}

	// this is a root
	if (GRAPH[v].tarjan_lowlink == GRAPH[v].tarjan_index) {

		// SCC has one element
		if (S.top() == v) {
			GRAPH[v].tarjan_inS = false;
			S.pop();
			// not a real SCC
			if (not(self_connected))
				GRAPH[v].to_be_deleted = true;

		// SCC has more elements
		} else {
			while (S.top() != v) {
				GRAPH[ S.top() ].tarjan_inS = false;
				S.pop();
			}
			S.pop();
			GRAPH[v].tarjan_inS = false;
		}

	}
}
#endif

#ifdef FERI_TARJANSEQUENTIAL
// find the current strongly connected component [Tarjan alg., Sequential Version]
void StrongConnect(GraphType &GRAPH, VertexStack &S, long int &index, VertexDescriptor v)
{
	// Variables for the sequential version
	VertexStack VStack;										// keep track of the active vertex
	SelfCStack SCStack;										// keep track if the active vertex has self connection
	outEdgeIteratorStack oEFirstStack, oESecondStack;		// keep track of the iterations on the edges
	bool in_action;											// keep track if the recursion goes one level down



	std::pair<outEdgeIterator, outEdgeIterator> outEdges;
	outEdgeIterator firstEdge, secondEdge, oute;

	VertexDescriptor w;


	// The active vertex is v
	VStack.push(v);

	// The first limit pair is the normal limit pair of v
	outEdges = boost::out_edges(v, GRAPH);
	oEFirstStack.push(outEdges.first);
	oESecondStack.push(outEdges.second);

	// Until we have vertex to check
	while ( VStack.size() > 0 ) {

		// current level of recursion
		in_action = true;

		// current vertex
		v = VStack.top();

		// getting limit pair of the edges
		firstEdge = oEFirstStack.top();
		secondEdge = oESecondStack.top();

		// the first edge to check
		oute = firstEdge;

		// first run for v
		if ( GRAPH[v].tarjan_index == -1 ) {

			SCStack.push(false);

			S.push(v);
			GRAPH[v].tarjan_inS = true;

			GRAPH[v].tarjan_index = index;
			GRAPH[v].tarjan_lowlink = index;

			++index;

		// we've just returned to v
		} else {

			// finish up the step after recursion towards w
			w = target(*oute, GRAPH);
			GRAPH[v].tarjan_lowlink = std::min(GRAPH[v].tarjan_lowlink, GRAPH[w].tarjan_lowlink);

			// step to next edge
			++oute;
		}

		// go through the edges determined by the limit pair, until we are on the same level of recursion
		while (( oute != secondEdge ) && (in_action)) {
			w = target(*oute, GRAPH);

			// there is a self-connection
			if (w == v) {
				SCStack.pop();
				SCStack.push(true);
			}

			// new node
			if (GRAPH[w].tarjan_index == -1) {

				// recursion level has changed
				in_action = false;

				// recursive call starting from w
				VStack.push(w);

				// we will continue from oute when we return to v
				oEFirstStack.pop();
				oEFirstStack.push( oute );

				// w starts with the default limit pair
				outEdges = boost::out_edges(w, GRAPH);
				oEFirstStack.push(outEdges.first);
				oESecondStack.push(outEdges.second);

			// not new node but already in S
			} else if (GRAPH[w].tarjan_inS) {

				GRAPH[v].tarjan_lowlink = std::min(GRAPH[v].tarjan_lowlink, GRAPH[w].tarjan_lowlink);

			}

			++oute;
		}

		// finish up if the recursion level has not changed
		if (in_action) {

			// this is a root
			if (GRAPH[v].tarjan_lowlink == GRAPH[v].tarjan_index) {

				// SCC has one element
				if (S.top() == v) {
					GRAPH[v].tarjan_inS = false;
					S.pop();
					// not a real SCC
					if (not( SCStack.top() ))
						GRAPH[v].to_be_deleted = true;

					// SCC has more elements
				} else {
					while (S.top() != v) {
						GRAPH[ S.top() ].tarjan_inS = false;
						S.pop();
					}
					S.pop();
					GRAPH[v].tarjan_inS = false;
				}
			}

			// pop everything related to this level
			SCStack.pop();
			oEFirstStack.pop();
			oESecondStack.pop();
			VStack.pop();

		}

	}
}
#endif

#if defined(FERI_TARJANSEQUENTIAL) || defined(FERI_TARJAN)
// check strongly connected components [Tarjan alg.]
void checkStronglyConnectedComponents(GraphType &GRAPH)
{
	VertexStack S;
	long int index = 0;

	std::pair<VertexIterator, VertexIterator> GRAPHlimits;
	VertexIterator it, it2;

	GRAPHlimits = boost::vertices(GRAPH);

	// reset properties
	it = GRAPHlimits.first;
	while ( it != GRAPHlimits.second ) {
		GRAPH[*it].to_be_deleted = false;
		GRAPH[*it].tarjan_inS = false;
		GRAPH[*it].tarjan_index = -1;
		++it;
	}

	// find SCC
	it = GRAPHlimits.first;
	while ( it != GRAPHlimits.second ) {

		if (GRAPH[*it].tarjan_index == -1) {
			StrongConnect(GRAPH, S, index, *it);
		}

		++it;

	}

	// delete nodes not in SCC
	it = GRAPHlimits.first;
	while ( it != GRAPHlimits.second ) {
		it2 = it;
		++it;
		if (GRAPH[*it2].to_be_deleted) {
			boost::clear_vertex(*it2, GRAPH);
			boost::remove_vertex(*it2, GRAPH);
		}
	}
}
#endif

// remove sources and sinks up to level max_iter (0 = unlimited)
void removeSourcesAndSinks(GraphType &GRAPH, const long int &max_iter)
{
	std::pair<VertexIterator, VertexIterator> GRAPHlimits;
	VertexIterator it, it2;

	long int iter;
	bool stop;



	iter = 0;
	do {

		stop = true;
		++iter;

		GRAPHlimits = boost::vertices(GRAPH);
		it = GRAPHlimits.first;

		while ( it != GRAPHlimits.second ) {

			it2 = it;
			++it;

			if (( boost::in_degree(*it2, GRAPH) == 0 ) || ( boost::out_degree(*it2, GRAPH) == 0 )) {
				stop = false;
				boost::clear_vertex(*it2, GRAPH);
				boost::remove_vertex(*it2, GRAPH);
			}

		}

	} while ( not(stop) && ((iter < max_iter) || (max_iter == 0)) );
}

// bisecting boxes
void bisectBoxes(GraphType &GRAPH)
{
	std::pair<VertexIterator, VertexIterator> GRAPHlimits;
	VertexIterator it, it2;

	std::vector<VertexDescriptor> n(2);
	VertexDescriptor vertex;

	VertexType vbox;
	VectorType box;

	std::pair<inEdgeIterator, inEdgeIterator> inEdges;
	inEdgeIterator ine;
	std::pair<outEdgeIterator, outEdgeIterator> outEdges;
	outEdgeIterator oute;

	int j;
	double maxdiam;



	GRAPHlimits = boost::vertices(GRAPH);
	it = GRAPHlimits.first;

	vbox.visited = true;

	while ( it != GRAPHlimits.second ) {

		if (not(GRAPH[*it].visited)) {

			box = GRAPH[*it].encl;

			j = box.dimension();
			maxdiam = diam(box(j)).leftBound();

			for (int i = box.dimension() - 1; i > 0; --i)
				if (diam(box(i)).leftBound() > maxdiam) {
					maxdiam = diam(box(i)).leftBound();
					j = i;
				}

			vbox.encl(j) = interval(box(j).leftBound(), box(j).mid().leftBound());
			for (int i = 1; i <= box.dimension(); ++i)
				if (i != j)
					vbox.encl(i) = box(i);
			n[0] = boost::add_vertex(vbox, GRAPH);

			vbox.encl(j) = interval(box(j).mid().leftBound(), box(j).rightBound());
			n[1] = boost::add_vertex(vbox, GRAPH);

			inEdges = boost::in_edges(*it, GRAPH);
			ine = inEdges.first;

			while ( ine != inEdges.second ) {
				vertex = source(*ine, GRAPH);

				for (int i = 1; i <= 2; ++i)
					boost::add_edge(vertex, n[i-1], GRAPH);
				++ine;
			}

			// self edges will be preserved
			outEdges = boost::out_edges(*it, GRAPH);
			oute = outEdges.first;

			while ( oute != outEdges.second ) {
				vertex = target(*oute, GRAPH);

				for (int i = 1; i <= 2; ++i)
					boost::add_edge(n[i-1], vertex, GRAPH);
				++oute;
			}

			it2 = it;
			++it;

			boost::clear_vertex(*it2, GRAPH);
			boost::remove_vertex(*it2, GRAPH);

		} else {

			++it;

		}

	}

	GRAPHlimits = boost::vertices(GRAPH);
	it = GRAPHlimits.first;

	while ( it != GRAPHlimits.second ) {
		GRAPH[*it].visited = false;
		++it;
	}

}

// keeping the good edges only
template<typename VFType>
void keepOnlyGoodEdges(VFType &TheMap, GraphType &GRAPH)
{
	std::pair<VertexIterator, VertexIterator> GRAPHlimits;
	VertexIterator it;

	VertexDescriptor vertex;

	VectorType box, img, tempbox;

	std::pair<outEdgeIterator, outEdgeIterator> outEdges;
	outEdgeIterator oute, oute2;



	GRAPHlimits = boost::vertices(GRAPH);
	it = GRAPHlimits.first;
	while ( it != GRAPHlimits.second ) {

		box = GRAPH[*it].encl;
		TheMap.evaluate(box, img);

		outEdges = boost::out_edges(*it, GRAPH);
		oute = outEdges.first;
		while ( oute != outEdges.second ) {
			vertex = target(*oute, GRAPH);

			if ( not(intersection(GRAPH[vertex].encl, img, tempbox)) ) {
				oute2 = oute;
				++oute;
				boost::remove_edge(oute2, GRAPH);
			} else {
				++oute;
			}
		}

		++it;
	}
}

// report the size of graph
void reportSizeOfGraph(const GraphType &GRAPH, std::ofstream &ostream)
{
	ostream << "  - Number of boxes: " << boost::num_vertices(GRAPH) << ", number of edges: " << boost::num_edges(GRAPH) << std::endl;
}

// save graph
template<typename VFType>
void saveGraph(VFType &TheMap, const GraphType &GRAPH, std::ofstream &ostream)
{
		std::pair<VertexIterator, VertexIterator> GRAPHlimits;
		VertexIterator it;

		VertexDescriptor vertex;

		VectorType box, img, tempbox;

		long int boxcount, edgecount;

		std::pair<outEdgeIterator, outEdgeIterator> outEdges;
		outEdgeIterator oute;

		GRAPHlimits = boost::vertices(GRAPH);
		it = GRAPHlimits.first;
		boxcount = 0;
		while ( it != GRAPHlimits.second ) {

			box = GRAPH[*it].encl;
			TheMap.evaluate(box, img);
			++boxcount;

			ostream << "Box no. " << boxcount << "  value: " << box << "  mapped to: " << img << std::endl;

			outEdges = boost::out_edges(*it, GRAPH);
			oute = outEdges.first;
			edgecount = 0;
			while ( oute != outEdges.second ) {
				++edgecount;
				vertex = target(*oute, GRAPH);
				ostream << "Edge no. " << boxcount << "/" << edgecount << " with target value: " << GRAPH[vertex].encl;
				if ( not(intersection(GRAPH[vertex].encl, img, tempbox)) )
					ostream << " and will be deleted.";
				ostream << std::endl;
				++oute;
			}
			++it;
		}
}

// check attraction to fixed point, true if all boxes are attracted
bool checkAttractionDomain(const VectorType &attDomain, GraphType &GRAPH)
{
	std::pair<VertexIterator, VertexIterator> GRAPHlimits;
	VertexIterator it, it2;

	VectorType box;



	bool result = true;

	#ifdef FERI_GRAPHICS_DRAW_EXTRA
		double lt1, lt2, rb1, rb2;
	#endif

	#ifdef FERI_GRAPHICS_DRAW_EXTRA

		lt1 = attDomain(1).leftBound();
		lt2 = attDomain(2).leftBound();
		rb1 = attDomain(1).rightBound();
		rb2 = attDomain(2).rightBound();

		rootFrame.box(lt1, lt2, rb1, rb2, YELLOW);
	#endif



	GRAPHlimits = boost::vertices(GRAPH);
	it = GRAPHlimits.first;

	while ( it != GRAPHlimits.second ) {

		box = GRAPH[*it].encl;

		if ( subset(box, attDomain) ) {

			#ifdef FERI_GRAPHICS_DRAW_EXTRA
				lt1 = box(1).leftBound();
				lt2 = box(2).leftBound();
				rb1 = box(1).rightBound();
				rb2 = box(2).rightBound();

				rootFrame.boxFill(lt1, lt2, rb1, rb2, GREEN);
			#endif

			it2 = it;
			++it;
			boost::clear_vertex(*it2, GRAPH);
			boost::remove_vertex(*it2, GRAPH);

		} else {

			result = false;
			++it;

		}

	}

	#ifdef FERI_GRAPHICS
		waitBt();
	#endif

	return result;
}

// ~ end of file ~
#endif
